""" Functions for loading and saving files to disk. """
from pathlib import Path

from mlspeclib.mlschema import MLSchema
from mlspeclib.helpers import convert_dict_to_yaml


class IO:
    """ Core IO class for reading and writing files."""
    @staticmethod
    def get_content_from_path(filepath: str) -> str:
        """ Takes a string, converts to a pathlib Path and loads the file as text. """
        return IO._load(filepath)

    @staticmethod
    def write_content_to_path(filepath: str, content):
        """ Takes a string, converts to a pathlib Path and writes the file as text. """
        if isinstance(content, dict):
            content_as_string = convert_dict_to_yaml(content)
        else:
            content_as_string = content
        return IO._save(filepath, content_as_string)

    # Moved to a function to allow swapping out for other libraries
    @staticmethod
    def _load(filepath: str) -> str:
        """ Internal load function, uses Path.read_text()."""
        return Path(filepath).read_text()

    # Moved to a function to allow swapping out for other libraries
    @staticmethod
    def _save(filepath: str, content: str):
        """ Writes file to disk using Path.write_text()."""
        return Path(filepath).write_text(content)



""" MLSchema object which converts yaml into objects and applies validation rules. """

import re
import os
from distutils import util

from yaml.scanner import ScannerError
from pathlib import Path

from marshmallow import Schema, fields, RAISE, validate, pre_load, ValidationError
import marshmallow.class_registry
from marshmallow.class_registry import RegistryError

import mlspeclib
from mlspeclib.helpers import (
    convert_yaml_to_dict,
    merge_two_dicts,
    contains_minimum_fields_for_schema,
    valid_comparison_operator,
    generate_lambda,
    build_schema_name_for_schema,
    build_schema_name_for_object,
)
from mlspeclib.mlschemafields import MLSchemaFields
from mlspeclib.mlschemavalidators import MLSchemaValidators

from ast import literal_eval

import logging


class MLSchema(Schema):
    """ Top level object for creating schema. Creates and stores schemas in
    marshmallow.class_registry (including nested schemas), and those schemas
    are used to .load({dicts}) into objects, and do schema verification."""

    schema_name = None

    # pylint: disable=too-few-public-methods
    class Meta:
        """ Meta options for MLSchema. """

        datetimeformat = "iso"  # ISO 8601 format
        render_module = "yaml"
        register = True
        unknown = RAISE

    # TODO: Want to add a DotMap to the underlying dictionary to make even cleaner access -
    # https://github.com/marshmallow-code/marshmallow/issues/1555
    # https://github.com/drgrib/dotmap

    # Functions below here are used for building schemas.

    # pylint: disable=unused-argument, protected-access
    @staticmethod
    def create_schema(raw_string: dict, schema_name: str = None):
        """ Uses create_schema_type to create a schema, and then instantiates it for return. """
        abstract_schema_type = MLSchema.create_schema_type(raw_string, schema_name)
        return abstract_schema_type()

    # pylint: disable=no-member, protected-access
    @staticmethod
    def create_schema_type(raw_string: dict, schema_name: str = None):
        """ Creates a new schema from a string of yaml. inheriting from MLSchema.\
            Schema still needs to be instantiated before use.

            e.g. this_schema = MLSchema.create_schema(raw_string)
                 schema = this_schema()
                 this_object = schema.load(object_submission_dict)
            """
        schema_as_dict = convert_yaml_to_dict(raw_string)
        schema_as_dict = MLSchema._augment_with_base_schema(schema_as_dict)

        if schema_name is None:
            # schema_name = foo_foo("a", "b")
            schema_name = build_schema_name_for_schema(
                mlspec_schema_version=schema_as_dict["mlspec_schema_version"],
                mlspec_schema_type=schema_as_dict["mlspec_schema_type"],
            )

        fields_dict = {}

        for field in schema_as_dict:
            if "marshmallow.fields" in str(
                type(schema_as_dict[field])
            ) or "mlschemafields.MLSchemaFields" in str(type(schema_as_dict[field])):
                # In this case, the field has already been created an instantiated properly (because
                # it comes from a base schema registered in marshmallow.class_registry). We can skip
                # all of the below and just add it to the field dict. This includes nested fields.
                fields_dict[field] = schema_as_dict[field]
            elif schema_as_dict[field] is None:
                raise AttributeError(
                    f"""It appears at the field '{field}' the yaml/dict \
                    is not formatted with attributes. Could it be an indentation error?"""
                )
            elif (
                "type" in schema_as_dict[field]
                and schema_as_dict[field]["type"].lower() == "nested"
            ):

                nested_schema_type = MLSchema.create_schema_type(
                    schema_as_dict[field]["schema"], schema_name + "_" + field.lower()
                )
                nested_schema_type.name = field
                fields_dict[field] = fields.Nested(nested_schema_type)
            else:
                field_method = MLSchema._field_method_from_dict(
                    field, schema_as_dict[field]
                )
                if field_method:
                    fields_dict[field] = field_method

        abstract_schema = MLSchema.from_dict(fields_dict)
        if schema_name:
            marshmallow.class_registry.register(schema_name, abstract_schema)
            abstract_schema.schema_name = schema_name
        return abstract_schema

    @staticmethod
    def _field_method_from_dict(name: str, field_dict: dict):
        """ Takes the dict from a yaml schema and creates a field appropriate for Marshmallow.  """
        field_types = {
            "string": fields.Str(),
            "uuid": fields.UUID(),
            "uri": fields.Str(validate=MLSchemaValidators.validate_type_URI),
            "datetime": MLSchemaFields.DateTime(),
            "semver": fields.Str(validate=MLSchemaValidators.validate_type_semver),
            "allowed_schema_types": fields.Str(),
            "boolean": fields.Boolean(),
            "list": fields.List(fields.Raw()),
            "list_strings": fields.List(
                fields.Str(validate=MLSchemaValidators.validate_type_string_cast)
            ),
            "list_of_tensor_shapes": fields.List(
                fields.Tuple([fields.Str(), fields.List(fields.Int)])
            ),
            "list_interfaces": fields.List(
                fields.Dict(validate=MLSchemaValidators.validate_type_interfaces)
            ),
            "workflow_steps": fields.Dict(
                validate=MLSchemaValidators.validate_type_workflow_steps
            ),
            "tags": fields.List(fields.Tuple([fields.Str(), fields.Str()])),
            "path": fields.Str(validate=MLSchemaValidators.validate_type_path),
            "dict": fields.Dict(),
            "float": fields.Float(),
            "email": fields.Email(),
            "bucket": fields.Str(validate=MLSchemaValidators.validate_type_bucket),
            "int": fields.Int(),
        }

        try:
            if "meta" in field_dict:
                # The field is a meta field about the schema, so skip adding a method
                return None
            field_type = field_dict["type"].lower()
            field_declaration = field_types[field_type]
        except KeyError:
            raise AttributeError(
                f"MLSchema Library has no field type named '{field_type}''"
            )

        # Need to put this first so that we can redeclare the field function. Tried
        # attaching the regex without using the fields.Str(validate=) format,
        # and it didn't seem to work.
        if "regex" in field_dict:
            try:
                re.compile(field_dict["regex"])
            except re.error:
                raise AssertionError(
                    f"The regex ('{field_dict['regex']}') does not appear to be a valid regex."
                )

            field_declaration = fields.Str(
                validate=validate.Regexp(
                    field_dict["regex"], error=f"No match for in field: {name}"
                )
            )

        if "allowed" in field_dict:
            # TODO: This may be a bug in waiting - would prefer not to overwrite, but instead
            # just to add. Filed a bug with marshmallow to see.
            # TODO: Bug - cannot currently support with "list"
            field_declaration = fields.Str(
                validate=validate.OneOf(field_dict["allowed"])
            )

        if "constraint" in field_dict:
            # TODO: There's probably a better place for this - it's validating
            # the constraint on an int or float that it's actually in the correct form but
            # isn't changing anything else.
            if field_type not in ["int", "float"]:
                raise ValueError(
                    "Attempting to add a 'constraint' to a field that does not appear to be a 'float' or 'int'."
                )
            if field_type == "int":
                field_declaration = fields.Int(
                    validate=generate_lambda(field_dict["constraint"])
                )
            elif field_type == "float":
                field_declaration = fields.Float(
                    validate=generate_lambda(field_dict["constraint"])
                )
            else:
                raise ValueError("Don't know how you got here.")

        if "required" in field_dict and util.strtobool(
            MLSchemaValidators.validate_bool_and_return_string(field_dict["required"])
        ):
            field_declaration.required = True
        else:
            field_declaration.allow_none = True

        if "empty" in field_dict and util.strtobool(
            MLSchemaValidators.validate_bool_and_return_string(field_dict["empty"])
        ):
            field_declaration.allow_none = True

        return field_declaration

    @staticmethod
    def _augment_with_base_schema(schema_dict: dict):
        # The below is incredibly gross code smell - ideally, there'd
        # be a much better way to merge the base schema with the derivative ones
        # and that allowed for multiple layers of inheritance. Oh well.
        # TODO: Support multiple layer of inheritence

        base_name = None
        base_version = None

        # If the yaml has a base_type
        if (
            "mlspec_base_type" in schema_dict
            and schema_dict["mlspec_base_type"]["meta"] is not None
        ):
            base_type = schema_dict["mlspec_base_type"]
        else:
            # There is no base type, so just return the full schema.
            return schema_dict

        if "mlspec_schema_version" in schema_dict:
            base_version = schema_dict["mlspec_schema_version"]
        else:
            raise KeyError(
                f"There is no mlschema version for this spec, so cannot look up the base schema."
            )  # noqa
        try:
            base_schema = marshmallow.class_registry.get_class(
                build_schema_name_for_schema(
                    mlspec_schema_version=base_version, mlspec_schema_type=base_type
                )
            )
        except RegistryError:
            raise RegistryError(
                f"""Could not find the base schema in the class \
        registry. Values provided:
        base_name = '{base_name}'
        schema_version = '{base_version}'"""
            )

        base_dict = base_schema().fields

        schema_dict = merge_two_dicts(schema_dict, base_dict)

        return schema_dict

    # Functions below here are used for loading objects

    # pylint: disable=unused-argument, protected-access
    @pre_load
    def pre_load_data(self, data, **kwargs):
        """ Pre_load accomplishes two things:
            - First it builds the schema_name according to a standard
            pattern (schema_version with '.' replaced with '_',
            then an '_' then the specific schema name.).

            This means a standard registered schema name will look like "0_0_1_datapath".

            It needs to build this standard schema name because we will use it to
            retrieve the schema from marshmallow.class_registry

            - Second, this function checks for any fields that are of type -
            fields.Nested() and, if so, it looks up the schemas for those fields, and
            converts the data in those fields into an object. We need to do this because
            marshmallow requires an object (not a dict) to be attached in order to do
            validation on nested fields."""

        # TODO: May be worth exploring if we could create a custom validation type that used
        # dicts, instead of subobjects, because that would simplify a LOT of this code.

        data = convert_yaml_to_dict(data)
        schema_name = build_schema_name_for_object(self, data)

        try:
            marshmallow.class_registry.get_class(schema_name)
        except AttributeError:
            raise AttributeError(f"{schema_name} is not a valid schema type.")

        return data

    @staticmethod
    def create_object(submission_text: str):
        """ Creates an object that can be read and written to. """
        submission_dict = convert_yaml_to_dict(submission_text)
        schema = MLSchema.load_schema_from_registry(data=submission_dict)
        return schema().load(submission_dict)

    @staticmethod
    def load_schema_from_registry(data: dict):
        """ Loads just the schema from marshmallow's class_registry using the
        data (schema_type and schema_version) loaded from the submission. """

        schema_name = build_schema_name_for_object(submission_data=data)
        return marshmallow.class_registry.get_class(schema_name)

    # Functions below here are for filling out the registry
    @staticmethod
    def populate_registry():
        """ Loads all the base schemas for the schema registry. """

        schemas_to_process = []
        no_base = []
        has_base = []
        last = []

        load_root = os.path.dirname(mlspeclib.__file__)
        # logging.debug(f"Registry load root: {load_root}")
        load_path = Path(load_root).glob("schemas/**/*.yaml")
        # logging.debug(f"Registry load path: {load_path}")
        load_list = list(load_path)
        # logging.debug(f"Registry load list: {load_list}")

        for schema_file in load_list:
            schema_text = schema_file.read_text()
            schema_dict = convert_yaml_to_dict(schema_text)

            if "last" in schema_dict["mlspec_schema_type"]:
                last.append(schema_dict)
            elif "mlspec_base_type" not in schema_dict:
                no_base.append(schema_dict)
            else:
                has_base.append(schema_dict)

        schemas_to_process = no_base + has_base + last

        for loaded_schema in schemas_to_process:
            loaded_schema_name = build_schema_name_for_schema(
                mlspec_schema_type=loaded_schema["mlspec_schema_type"],
                mlspec_schema_version=loaded_schema["mlspec_schema_version"],
            )
            try:
                marshmallow.class_registry.get_class(loaded_schema_name)
            except RegistryError:
                MLSchema.create_schema(loaded_schema)

    @staticmethod
    def append_schema_to_registry(load_path: Path) -> bool:
        if not isinstance(load_path, Path):
            raise TypeError("Appending schemas to a registry expects a Path object.")

        all_found_files = list(load_path.glob("**/*.yaml"))

        if len(all_found_files) == 0:
            raise FileNotFoundError(
                f"No files ending in '.yaml' were found in the path '{load_path}'"
            )

        all_schemas = []
        files_with_errors = []

        for putative_schema_file in all_found_files:
            this_text = putative_schema_file.read_text()
            try:
                this_dict = convert_yaml_to_dict(this_text)
            except ScannerError as se:
                files_with_errors.append(
                    (
                        putative_schema_file.name,
                        f"Yaml could not be parsed. Error details: \n{str(se)}",
                    )
                )
                continue

            if not contains_minimum_fields_for_schema(this_dict):
                files_with_errors.append(
                    (
                        putative_schema_file.name,
                        f"""Does not contain all of the minimum schema necessary as top level fields - list includes: mlspec_schema_version, mlspec_schema_version.meta, mlspec_base_type, mlspec_base_type.meta, mlspec_schema_type, mlspec_schema_type.meta, schema_version and schema_type""",
                    )
                )
                continue

            all_schemas.append(this_dict)

        if len(files_with_errors) > 0:
            error_string = ""
            for err in files_with_errors:
                error_string += f"{err[0]}: {err[1]}\n"

            print(error_string)
            return False

        for schema_dict in all_schemas:
            schema_name = build_schema_name_for_schema(
                mlspec_schema_type=schema_dict["mlspec_schema_type"],
                mlspec_schema_version=schema_dict["mlspec_schema_version"],
            )
            try:
                marshmallow.class_registry.get_class(schema_name)
            except RegistryError:
                MLSchema.create_schema(schema_dict)



import sys
import yaml
from pathlib import Path

sys.path.append("..")
sys.path.append(Path(__file__).parent.resolve())

from mlspeclib.mlschema import MLSchema
from mlspeclib.mlschemaenums import MLSchemaTypes
from mlspeclib.mlobject import MLObject
from mlspeclib.helpers import return_schema_name, convert_yaml_to_dict
import tempfile

import json
import uuid
import logging

from azure.cosmos import exceptions, CosmosClient, PartitionKey
import traceback

logging.basicConfig(level=logging.DEBUG)

class CosmosHelpers:
    edges_to_submit = {}
    _url = None
    _key = None
    _database_name = None
    _container_name = None
    _workflow_partition_id = None
    _cosmos_client = None
    _database_client = None
    _workflow_node_id = None

    def __init__(
        self,
        url=None,
        key=None,
        database_name=None,
        container_name=None,
        credential_dict: str = None,
    ):
        if credential_dict is not None:
            url = credential_dict["url"]
            key = credential_dict["key"]
            database_name = credential_dict["database_name"]
            container_name = credential_dict["container_name"]

        self._url = url
        self._key = key
        self._database_name = database_name
        self._container_name = container_name

        self._workflow_id = uuid.uuid4()

        self._cosmos_client = CosmosClient(self._url, self._key)
        self._db_client = self._cosmos_client.get_database_client(f"/dbs/{database_name}")
        self._container_client = self._db_client.get_container_client(f"/colls/{container_name}")

    # def cleanup_graph(self):
    #     logging.debug(
    #         "\tRunning this Gremlin query:\n\t{0}".format(self._gremlin_cleanup_graph)
    #     )
    #     callback = self._gremlin_client.submitAsync(self._gremlin_cleanup_graph)
    #     if callback.result() is not None:
    #         logging.debug("\tCleaned up the graph!")

    # def build_schema_name(self, workflow_step_object):
    #     return return_schema_name(
    #         workflow_step_object["schema_version"], workflow_step_object["schema_type"]
    #     )

    # def create_workflow_node(self, schema_version):
    #     query = f"g.addV('workflow').property('id', '{schema_version}').property('workflow_id', '{self._workflow_id}')"
    #     returned_node = self.execute_query(query, True)
    #     self._workflow_node_id = returned_node["id"]

    # def insert_vertex(
    #     self,
    #     step_name,
    #     in_schema,
    #     execution_schema,
    #     out_schema,
    #     workflow_id,
    #     previous_step=None,
    #     next_step=None,
    # ):

    #     insert_query = f"""g.addV('step').property('id', '{step_name}')
    #             .property('in_schema', '{self.__build_schema_name(in_schema)}')
    #             .property('execution_schema', '{self.__build_schema_name(execution_schema)}')
    #             .property('out_schema', '{self.__build_schema_name(out_schema)}')
    #             .property('workflow_id', '{workflow_id}')"""

    #     self.edges_to_submit[step_name] = (previous_step, next_step)

    #     self.execute_query(insert_query)

    #     part_of_query = (
    #         f"g.V('{step_name}').addE('part_of').to(g.V('{self._workflow_node_id}'))"
    #     )
    #     logging.debug(f"Part_of_query: {part_of_query}")
    #     self.execute_query(part_of_query)

    #     contains_query = (
    #         f"g.V('{step_name}').addE('contains').from(g.V('{self._workflow_node_id}'))"
    #     )
    #     logging.debug(f"Contains_query: {contains_query}")
    #     self.execute_query(contains_query)

    # def attach_step_info(
    #     self, mlobject: MLObject, workflow_id, step_name: str, step_type: str
    # ):
    #     if step_type not in ["input", "execution", "output"]:
    #         raise ValueError(
    #             f"Error when saving '{mlobject.get_schema_name()}', the content_type must be from ['input', 'execution', 'output']."
    #         )
    #     run_info_id = f"{step_name}|{mlobject.run_id}|{mlobject.run_date}"
    #     add_run_info_query = f"""g.addV('step_run').property('id', '{run_info_id}')
    #                         .property('run_id', '{mlobject.run_id}')
    #                         .property('run_date', '{mlobject.run_date}')
    #                         .property('{step_type}_content', '{mlobject.to_yaml()}')"""

    #     self.execute_query(add_run_info_query)

    #     self.execute_query(
    #         f"g.V('{self._workflow_node_id}').out().hasId({step_name}).addE('results').to(g.V('{run_info_id}'))"
    #     )
    #     self.execute_query(
    #         f"g.V('{self._workflow_node_id}').out().hasId({step_name}).addE('root').from(g.V('{run_info_id}'))"
    #     )

    # def connect_workflow_steps(self):
    #     for edge in self.edges_to_submit:
    #         self.connect_vertices(
    #             edge, self.edges_to_submit[edge][0], self.edges_to_submit[edge][1]
    #         )

    #     self.edges_to_submit = {}

    # def connect_next_previous_vertices(self, step_name, previous_step, next_step):
    #     if previous_step is not None:
    #         previous_query = (
    #             f"g.V('{step_name}').addE('previous').from(g.V('{previous_step}'))"
    #         )
    #         self.execute_query(previous_query)

    #     if next_step is not None:
    #         next_query = f"g.V('{step_name}').addE('previous').from(g.V('{next_step}'))"
    #         self.execute_query(next_query)

    # def execute_query(self, query, return_node=False):
    #     logging.debug(f"Inside the execute query: {query}")

    #     # TODO: Need to implement sanitization.
    #     logging.warn(
    #         f"*** execute_query DOES NO SANITIZATION OF INPUTS. BE EXTREMELY CAREFUL. ***"
    #     )
    #     callback = self._gremlin_client.submitAsync(query)
    #     result = None
    #     if callback.result() is not None:
    #         result = callback.result().one()
    #         logging.debug("\tExecuted:\n\t{0}\n".format(result))
    #     else:
    #         logging.debug("Something went wrong with this query: {0}".format(query))

    #     if return_node:
    #         return result[0]

    # def __build_schema_name(self, workflow_step_object):
    #     return return_schema_name(
    #         workflow_step_object["schema_version"], workflow_step_object["schema_type"]
    #     )

    # def empty_graph(self):
    #     self.cleanup_graph()



import sys
import yaml
from pathlib import Path

from mlspeclib.mlschema import MLSchema
from mlspeclib.mlschemaenums import MLSchemaTypes
from mlspeclib.mlobject import MLObject
from mlspeclib.helpers import (
    return_schema_name,
    convert_yaml_to_dict,
    decode_raw_object_from_db,
)
from mlspeclib.experimental.gremlin_helpers import GremlinHelpers
import tempfile

import json
import uuid

import traceback

import logging


class Metastore:
    _gc = None

    def __init__(self, credentials):
        self._gc = GremlinHelpers(credential_dict=credentials)

    def save(self, mlobject: MLObject, workflow_version_id, step_name, content_type):
        """ Saves an MLObject to the metastore connection. Uses the run_id from the object,
        and attaches all non-internal fields on the MLObject as properties, as well as storing
        an encoded representation of the full yaml under raw_content.

        Also creates edges with the specific step as:
            step_name -> .out('result') -> mlobject
            mlobject -> .out('root') -> step_name"""

        self._gc.attach_step_info(
            mlobject, workflow_version_id, step_name, content_type
        )

    def get_all_runs(self, workflow_version_id, step_name):
        return self._gc.get_all_runs(workflow_version_id, step_name)

    def load(self, workflow_version, step_name, run_info_id):
        return self._gc.get_run_info(workflow_version, step_name, run_info_id)

    def get_workflow_object(self, workflow_version_id):
        full_results = self._gc.get_workflow_node(workflow_version_id)
        if len(full_results) == 0:
            return None
        elif len(full_results) > 1:
            logging.debug(
                "More than one workflow associated with this schema. Returning None."
            )
            return None
        else:
            raw_item = full_results[0]

        if (
            "properties" not in raw_item
            or "raw_content" not in raw_item["properties"]
            or len(raw_item["properties"]["raw_content"]) == 0
            or "value" not in raw_item["properties"]["raw_content"][0]
        ):
            return None

        obf_string = raw_item["properties"]["raw_content"][0]["value"]
        return MLObject.create_object_from_string(decode_raw_object_from_db(obf_string))

    def get_step_object(self, workflow_version_id, step_name, run_info_id):
        raw_item = self.load(workflow_version_id, step_name, run_info_id)

        if (
            "properties" not in raw_item
            or "raw_content" not in raw_item["properties"]
            or len(raw_item["properties"]["raw_content"]) == 0
            or "value" not in raw_item["properties"]["raw_content"][0]
        ):
            return None

        obf_string = raw_item["properties"]["raw_content"][0]["value"]
        return MLObject.create_object_from_string(decode_raw_object_from_db(obf_string))

    def empty_graph(self):
        return self._gc.empty_graph()

    def create_workflow_node(self, workflow_object: MLObject):
        return self._gc.create_workflow_node(
            workflow_object.get_schema_version(), workflow_object
        )

    def create_workflow_steps(self, raw_workflow_node, workflow_object: MLObject):
        edges_to_submit = {}
        for step_name in workflow_object.steps:
            step_contents = workflow_object.steps[step_name]

            next_step = None
            previous_step = None
            if "previous" in step_contents:
                previous_step = step_contents.previous
            if "next" in step_contents:
                next_step = step_contents.next

            step_name = step_name
            input_schema_version = step_contents.input.schema_version
            input_schema_type = step_contents.input.schema_type
            execution_schema_version = step_contents.execution.schema_version
            execution_schema_type = step_contents.execution.schema_type
            output_schema_version = step_contents.output.schema_version
            output_schema_type = step_contents.output.schema_type

            workflow_version_id = raw_workflow_node[0]["properties"]["version"][0][
                "value"
            ]
            previous_step = previous_step
            next_step = next_step

            edges_to_submit[step_name] = (previous_step, next_step)

            self._gc.insert_workflow_step(
                step_name=step_name,
                input_schema_version=input_schema_version,
                input_schema_type=input_schema_type,
                execution_schema_version=execution_schema_version,
                execution_schema_type=execution_schema_type,
                output_schema_version=output_schema_version,
                output_schema_type=output_schema_type,
                workflow_version_id=workflow_version_id,
                previous_step=previous_step,
                next_step=next_step,
            )

        self._gc.connect_workflow_steps(edges_to_submit)



import sys
import yaml
from pathlib import Path

from mlspeclib.mlschema import MLSchema
from mlspeclib.mlschemaenums import MLSchemaTypes
from mlspeclib.mlobject import MLObject
from mlspeclib.helpers import (
    return_schema_name,
    convert_yaml_to_dict,
    to_yaml,
    to_json,
    encode_raw_object_for_db,
)
from collections import OrderedDict
import tempfile

import json
import uuid
import logging

from gremlin_python import statics
from gremlin_python.structure.graph import Graph
from gremlin_python.process.graph_traversal import __
from gremlin_python.process.strategies import *
from gremlin_python.driver.driver_remote_connection import DriverRemoteConnection
from gremlin_python.driver import client, serializer, resultset

import asyncio

import pymysql

import traceback

import tornado

sys.path.append("..")
sys.path.append(Path(__file__).parent.resolve())


class GremlinHelpers:
    _gremlin_cleanup_graph = "g.V().drop()"
    # _gremlin_cleanup_graph = ""
    _gremlin_count_vertices = "g.V().count()"

    _url = None
    _key = None
    _database_name = None
    _container_name = None
    _workflow_partition_id = None
    _gremlin_client = None
    _workflow_node_id = None
    _request = None
    _driver_remote_connection = None
    _graph = None

    def __init__(
        self,
        url=None,
        key=None,
        database_name=None,
        container_name=None,
        credential_dict: dict = None,
    ):
        if credential_dict is not None:
            url = credential_dict["url"]
            key = credential_dict["key"]
            database_name = credential_dict["database_name"]
            container_name = credential_dict["container_name"]

        self._url = url
        self._key = key
        self._database_name = database_name
        self._container_name = container_name

        self._workflow_partition_id = uuid.uuid4()

        print(self._url)
        self._gremlin_client = client.Client(
            f"{self._url}",
            "g",
            username=f"/dbs/{self._database_name}/colls/{self._container_name}",
            password=f"{self._key}",
            message_serializer=serializer.GraphSONSerializersV2d0(),
        )
        # self._request = httpclient.HTTPRequest(f"{self._url}", headers={"Authorization": "Token AZX ..."})
        # self._driver_remote_connection = DriverRemoteConnection(self._request.url, 'g', username=f"/dbs/{self._database_name}/colls/{self._container_name}", password=f"{self._key}",message_serializer=serializer.GraphSONSerializersV2d0())
        # self._graph = Graph()
        # self._g = self._graph.traversal().withRemote(self._driver_remote_connection, 'g')

    def cleanup_graph(self):
        logging.debug(
            "\tRunning this Gremlin query:\n\t{0}".format(self._gremlin_cleanup_graph)
        )
        callback = self._gremlin_client.submitAsync(self._gremlin_cleanup_graph)
        if callback.result() is not None:
            logging.debug("\tCleaned up the graph!")

    def build_schema_name(self, workflow_step_object):
        return return_schema_name(
            workflow_step_object["schema_version"], workflow_step_object["schema_type"]
        )

    def create_workflow_node(self, workflow_version, mlobject):
        # mlobject_dict = mlobject.dict_without_internal_variables()
        # property_string = convert_to_property_strings(mlobject_dict)
        raw_content = encode_raw_object_for_db(mlobject)
        save_workflow_query = f"g.addV('id', 'workflow').property('version', '{workflow_version}').property('workflow_partition_id', '{self._workflow_partition_id}').property('raw_content', '{raw_content}')"

        return self.execute_query(save_workflow_query)

    def get_workflow_node(self, schema_version):
        query = "g.V().has('id', 'workflow').has('version', '%s')"
        return self.execute_query(sQuery(query, [schema_version]))

    def insert_workflow_step(
        self,
        step_name,
        input_schema_version,
        input_schema_type,
        execution_schema_version,
        execution_schema_type,
        output_schema_version,
        output_schema_type,
        workflow_version_id,
        previous_step=None,
        next_step=None,
    ):

        params = [
            step_name,
            input_schema_version,
            input_schema_type,
            execution_schema_version,
            execution_schema_type,
            output_schema_version,
            output_schema_type,
            workflow_version_id,
            str(self._workflow_partition_id),
        ]

        insert_query = sQuery(
            f"""g.addV('id','%s').property('type', 'workflow_step').property('input_schema_version', '%s').property('input_schema_type', '%s').property('execution_schema_version', '%s').property('execution_schema_type', '%s').property('output_schema_version', '%s').property('output_schema_type', '%s').property('workflow_version_id', '%s').property('workflow_partition_id', '%s')""",
            params,
        )

        self.execute_query(insert_query)

        part_of_query = f"g.V('{step_name}').addE('part_of').to(g.V().has('id', 'workflow').has('version', '{workflow_version_id}'))"
        logging.debug(f"Part_of_query: {part_of_query}")
        self.execute_query(part_of_query)

        contains_query = f"g.V('{step_name}').addE('contains').from(g.V().has('id', 'workflow').has('version', '{workflow_version_id}'))"
        logging.debug(f"Contains_query: {contains_query}")
        self.execute_query(contains_query)

    def attach_step_info(
        self, mlobject: MLObject, workflow_version_id, step_name: str, step_type: str
    ):
        if step_type not in ["input", "execution", "output", "log"]:
            raise ValueError(
                f"Error when saving '{mlobject.get_schema_name()}', the step_type must be from ['input', 'execution', 'output', 'log']."
            )
        run_info_id = get_run_info_id(step_name, mlobject.run_id, mlobject.run_date)
        mlobject_dict = mlobject.dict_without_internal_variables()
        property_string = convert_to_property_strings(mlobject_dict)
        raw_content = encode_raw_object_for_db(mlobject)
        add_run_info_query = f"""g.addV('id', '{run_info_id}'){property_string}.property('raw_content', '{raw_content}').property('workflow_partition_id', '{self._workflow_partition_id}')"""

        self.execute_query(add_run_info_query)

        self.execute_query(
            sQuery(
                "g.V('id', 'workflow').has('version', '%s').out().hasId('%s').addE('results').to(g.V('%s')).executionProfile()",
                [workflow_version_id, step_name, run_info_id],
            )
        )
        self.execute_query(
            sQuery(
                "g.V('id', 'workflow').has('version', '%s').out().hasId('%s').addE('root').from(g.V('%s')).executionProfile()",
                [workflow_version_id, step_name, run_info_id],
            )
        )

    def connect_workflow_steps(self, edges_to_submit):
        for edge in edges_to_submit:
            self.connect_next_previous_vertices(
                edge, edges_to_submit[edge][0], edges_to_submit[edge][1]
            )

    def connect_next_previous_vertices(self, step_name, previous_step, next_step):
        if previous_step is not None:
            previous_query = sQuery(
                f"g.V('%s').addE('previous').from(g.V('%s'))",
                [step_name, previous_step],
            )
            self.execute_query(previous_query)

        if next_step is not None:
            next_query = sQuery(
                f"g.V('%s').addE('next').to(g.V('%s'))", [step_name, next_step]
            )

            self.execute_query(next_query)

    def execute_query(self, query):
        logging.debug(f"Inside the execute query: {query}")

        # TODO: Need to implement sanitization.
        logging.warn(
            f"*** execute_query DOES NO SANITIZATION OF INPUTS. BE EXTREMELY CAREFUL. ***"
        )
        callback = self._gremlin_client.submitAsync(query)
        collected_result = []
        try:
            if callback.result() is not None:
                while True:
                    result_set = callback.result()
                    for result in result_set:
                        collected_result.extend(result)

                    if callback.done():
                        break

                    loop_sleep()

                logging.debug("\tExecuted:\n\t{0}\n".format(collected_result))
        except tornado.iostream.StreamClosedError as sce:
            logging.debug("Something went wrong with this query: {0}".format(query))
            logging.debug(f"Full error here: {str(sce)}")

        if len(collected_result) == 0:
            raise Exception(msg="Query returned zero results.")

        return collected_result

    def get_all_runs(
        self, workflow_version_id, step_name, descending_order=True
    ) -> list:
        order = "decr"
        if not descending_order:
            order = "incr"

        query = "g.V('id','workflow').has('version', '%s').out().has('id', '%s').out('results').order().by('run_date', %s)"

        results = self.execute_query(
            sQuery(query, [workflow_version_id, step_name, order])
        )

        result_dict = OrderedDict()
        for result in results:
            result_dict[result["id"]] = result

        return result_dict

    def get_run_info(self, workflow_version_id, step_name, run_info_id) -> dict:
        all_results = self.get_all_runs(
            workflow_version_id=workflow_version_id, step_name=step_name
        )
        if len(all_results) > 0:
            return all_results[run_info_id]
        else:
            return None

    def __build_schema_name(self, workflow_step_object):
        return return_schema_name(
            workflow_step_object["schema_version"], workflow_step_object["schema_type"]
        )

    def empty_graph(self):
        self.cleanup_graph()


async def loop_sleep():
    await asyncio.sleep(0.2)


def convert_to_property_strings(this_dict: dict, prefix=None):
    return_string = ""
    for key in this_dict:
        if isinstance(this_dict[key], dict):
            convert_to_property_strings(this_dict[key], prefix=key)

        key_string = str(key)

        if prefix is not None:
            key_string = f"{prefix}.{key_string}"

        return_string += f".property('{pymysql.escape_string(key_string)}', '{pymysql.escape_string(str(this_dict[key]))}')"
        # return_string += f".property({MySQLdb.escape_string(key_string,'utf-8')}, {MySQLdb.escape_string(str(this_dict[key]),'utf-8')})"

    return return_string


def sQuery(query, parameters: list = []):
    num_of_params = query.count("%s")
    if len(parameters) != num_of_params:
        raise ValueError(
            f"Number of parameters ({num_of_params}) for query '{query}' not equal to parameters. Parameters given: {parameters}"
        )

    safe_parameters = []
    for param in parameters:
        safe_parameters.append(pymysql.escape_string(param))

    return query % tuple(safe_parameters)


def get_run_info_id(step_name, run_id, run_date):
    return f"{step_name}|{run_id}|{run_date}"



# -*- coding: utf-8 -*-

# Learn more: https://github.com/kennethreitz/setup.py

from setuptools import setup, find_packages
import os

with open("README.md") as f:
    readme = f.read()


def package_files():
    paths = []
    for (path, directories, filenames) in os.walk("mlspeclib/schemas"):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


extra_files = package_files()

setup(
    name="mlspeclib",
    version="0.0.3",
    description="MLSpec helper library to making using metadata in ML workflows easier",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="David Aronchick",
    author_email="aronchick@gmail.com",
    url="https://github.com/mlspec/mlspec-lib",
    license="MIT License",
    keywords=["MLSpec", "Machine Learning"],  # Keywords that define your package best
    install_requires=[
        "wheel",
        "pyyaml",
        "marshmallow",
        "uritools",
        "semver",
        "python-box",
        "gremlinpython",
        "pymysql",
    ],
    packages=["mlspeclib", "mlspeclib.experimental"],
    include_package_data=True,
    package_data={"": extra_files},
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",  # Define that your audience are developers
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",  # Again, pick a license
        "Programming Language :: Python :: 3",  # Specify which pyhton versions that you want to support
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



""" Functions for validating submissions """
from distutils import util
import uritools
import semver as sv
import re
import marshmallow
from marshmallow import ValidationError
from marshmallow.class_registry import RegistryError
from mlspeclib.helpers import return_schema_name


# pylint: disable=missing-class-docstring
class MLSchemaValidators:
    @staticmethod
    def validate_type_semver(value):
        """ Uses the semver library to validate Semantic Version. Returns True/False """
        return sv.VersionInfo.isvalid(value)

    # pylint: disable=invalid-name
    @staticmethod
    def validate_type_URI(value):
        """ Uses the distutils library to validate date time. Returns True/False """
        return uritools.isuri(value)

    @staticmethod
    def validate_type_path(value):
        """ Uses regex to validate the value is a path. Returns True/False """
        path_regex = re.compile("(^[a-z0-9\-._~%!$&'()*+,;=:@/]+$)")  # noqa
        return path_regex.match(value)

    @staticmethod
    def validate_type_bucket(value):
        """ Uses regex to validate the value is a path. Returns True/False """
        bucket_regex = re.compile(
            "(?=^.{3,63}$)(?!^(\d+\.)+\d+$)(^(([a-z0-9]|[a-z0-9][a-z0-9\-]*[a-z0-9])\.)*([a-z0-9]|[a-z0-9][a-z0-9\-]*[a-z0-9])$)"
        )  # noqa
        return bucket_regex.match(value)

    @staticmethod
    def validate_type_string_cast(value):
        """ Casts value to string and validates. Returns True/False """
        # print(value)
        return isinstance(str(value), str)

    @staticmethod
    def validate_bool_and_return_string(val):
        """ Takes bool or string and converts value to string, testing
        if the result is true or false, and then casts result back to a string.

        Throws ValueError if it cannot be parsed as True/False.
        """

        # Convert to string first because some yaml libraries cast to booleans
        # and some leave as strings and we want to support both
        putative_bool_as_string = str(val)
        try:
            validated_bool = util.strtobool(putative_bool_as_string)
        except ValueError:
            raise ValueError(f"'{val}' is not a boolean value.")

        # We need to convert to 'True' or 'False' so first we cast to bool,
        # then to string.
        return str(bool(validated_bool))

    @staticmethod
    def validate_type_interfaces(interface_object: dict):
        """ Takes a Kubeflow Component interface and returns True/False if valid.
            From here: https://www.kubeflow.org/docs/pipelines/reference/component-spec/#detailed-specification-componentspec # noqa
        """
        # Kubeflow types manually copied from here -
        # https://github.com/kubeflow/pipelines/blob/master/sdk/python/kfp/dsl/types.py
        # TODO: Code gen this list: https://github.com/kubeflow/pipelines/blob/master/sdk/python/kfp/components/_structures.py # noqa
        kubeflow_types = {
            "Integer": int,
            "String": str,
            "Float": float,
            "Bool": bool,
            "List": list,
            "Dict": dict,
            "GcsPath": None,
            "GCPPath": None,
            "GCRPath": None,
            "GCPRegion": None,
            "GCPProjectID": None,
            "LocalPath": None,
            "Dataset": None,
            "AzureSku": None,
        }

        # The schema gives us a list of dicts, each of which only has one entry. So we have to do
        # this - could theoretically  support multiple entries per list item if it comes to that.

        if "name" in interface_object:
            # This is for the following:
            #   input:
            #   - { name: foo, type: bar}
            # TODO: THIS IS SUPER BROKEN, IT WILL FAIL ON INTERFACES NAMED 'name'
            interface_dict = interface_object
        else:
            # This is for the following:
            #   input:
            #   - foo: { type: bar}
            key = next(iter(interface_object.keys()))
            interface_dict = interface_object[key]
            interface_dict["name"] = key

        if "type" in interface_dict:
            # TODO: Let's raise this with KFP folks - I think type should be required.
            # if 'type' in interface_dict:
            #   raise ValidationError(f"No type given for interface.")

            if (
                isinstance(interface_dict["type"], dict)
                and len(interface_dict["type"].keys()) == 1
            ):
                interface_type = next(iter(interface_dict["type"].keys()))
            elif isinstance(interface_dict["type"], str):
                interface_type = interface_dict["type"]
            else:
                raise ValidationError(
                    f"{interface_dict['type']} is expected to be a string or a dict with one entry."
                )  # noqa

            if interface_type not in kubeflow_types.keys():
                raise ValidationError(
                    f"'{interface_type}' is not a known type for an interface. Types are case sensistive. Please see this link for known types: https://aka.ms/kfptypes."
                )  # noqa
            elif (
                "default" in interface_dict
                and kubeflow_types[interface_type] is not None
            ):
                try:
                    kubeflow_types[interface_type](interface_dict["default"])
                except (ValueError, TypeError):
                    raise ValidationError(
                        f"'{interface_dict['default']}' is not a valid default type for this field, nor is it castable using the provided type. If you were expecting it to be a string, make sure it's quoted."
                    )  # noqa

        # print(f"dict: f'{interface_dict}'")
        # print(f"key: '{key}'\ndict: f'{interface_dict}'")

        if "name" in interface_dict and not isinstance(interface_dict["name"], str):
            raise ValidationError(
                f"{interface_dict['name']} is not a valid string (if you were expecting it to be cast as a string make sure it's quoted."
            )  # noqa

        if "description" in interface_dict and not isinstance(
            interface_dict["description"], str
        ):
            raise ValidationError(
                f"{interface_dict['description']} is not a valid string (if you were expecting it to be cast as a string make sure it's quoted."
            )  # noqa

        return True

    @staticmethod
    def validate_type_workflow_steps(workflow_step: dict, **kwargs):
        required_fields = ["input", "execution", "output"]
        required_properties = ["schema_type", "schema_version"]
        next_previous_fields = ["next", "previous"]
        step_name = next(iter(workflow_step))
        step_contents = workflow_step[step_name]

        # if len(workflow_step) > 1:
        #     raise ValidationError(
        #         f"Somehow the workflow step '{step_name}' is has more than one entry. I don't know how. Sorry."
        #     )

        for field in next_previous_fields:
            if field in step_contents:
                if not isinstance(step_contents[field], str):
                    raise ValidationError(
                        f"The '{field}' step should be a string identifying another step in this workflow."
                    )
                # TODO: is it a listed step? (Can't do right now?) Here we're given a step, not an entire workflow

        for field in required_fields:
            if field in next_previous_fields:
                continue

            if field not in step_contents.keys():
                raise ValidationError(
                    f"There was no '{field}' in the step '{step_name}'."
                )

            for prop in required_properties:
                if prop not in step_contents[field]:
                    raise ValidationError(
                        f"There was no '{prop}' field for '{step_name}.{field}'."
                    )

            if not MLSchemaValidators.validate_type_semver(
                step_contents[field]["schema_version"]
            ):
                raise ValidationError(
                    f"The semver '{step_contents[field]['schema_version']}' is '{step_name}.{field}' is not a valid semver."
                )

            schema_name = return_schema_name(
                raw_schema_type_string=step_contents[field]["schema_type"],
                raw_schema_version_string=step_contents[field]["schema_version"],
            )

            try:
                marshmallow.class_registry.get_class(schema_name)
            except RegistryError:
                raise ValidationError(
                    f"Could not find the schema '{schema_name}' in the class registry."
                )

        return True
